// Copyright 2025 The Jujutsu Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package default_index;

// Control file to link operation to index segments.
message SegmentControl {
  // Hash (or file name) of the commit index segment.
  bytes commit_segment_id = 1;
  // First commit stored in the changed-path index segments. Unset if
  // changed-path index is disabled.
  optional uint32 changed_path_start_commit_pos = 2;
  // Hashes (or file names) of the changed-path index segments.
  repeated bytes changed_path_segment_ids = 3;
}
