// Copyright 2023 The Jujutsu Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package git_store;

message Commit {
  repeated bytes predecessors = 2;
  bytes change_id = 4;

  // Alternating positive and negative terms. Set only for conflicts.
  // Resolved trees are stored in the git commit
  repeated bytes root_tree = 1;
  // TODO(#1624): delete when we assume that all commits use this format
  bool uses_tree_conflict_format = 10;

  bool is_open = 8 [deprecated = true];
  bool is_pruned = 9 [deprecated = true];
}
