// This file is @generated by prost-build.
pub mod default_index {
    include!("default_index.rs");
}
pub mod git_store {
    include!("git_store.rs");
}
pub mod local_working_copy {
    include!("local_working_copy.rs");
}
pub mod secure_config {
    include!("secure_config.rs");
}
pub mod simple_op_store {
    include!("simple_op_store.rs");
}
pub mod simple_store {
    include!("simple_store.rs");
}
pub mod simple_workspace_store {
    include!("simple_workspace_store.rs");
}
