// This file is @generated by prost-build.
/// Metadata about the configuration files for jj
#[derive(Clone, PartialEq, Eq, Hash, ::prost::Message)]
pub struct ConfigMetadata {
    /// The absolute path to the workspace's .jj or .jj/repo directory for
    /// workspaces and repos respectively.
    /// Access via file_util::path_to/from_bytes
    #[prost(bytes = "vec", optional, tag = "1")]
    pub path: ::core::option::Option<::prost::alloc::vec::Vec<u8>>,
}
