// Copyright 2025 The Jujutsu Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package secure_config;

// Metadata about the configuration files for jj
message ConfigMetadata {
  // The absolute path to the workspace's .jj or .jj/repo directory for
  // workspaces and repos respectively.
  // Access via file_util::path_to/from_bytes
  optional bytes path = 1;
}