// This file is @generated by prost-build.
#[derive(Clone, PartialEq, Eq, Hash, ::prost::Message)]
pub struct TreeValue {
    #[prost(oneof = "tree_value::Value", tags = "2, 3, 4")]
    pub value: ::core::option::Option<tree_value::Value>,
}
/// Nested message and enum types in `TreeValue`.
pub mod tree_value {
    #[derive(Clone, PartialEq, Eq, Hash, ::prost::Message)]
    pub struct File {
        #[prost(bytes = "vec", tag = "1")]
        pub id: ::prost::alloc::vec::Vec<u8>,
        #[prost(bool, tag = "2")]
        pub executable: bool,
        #[prost(bytes = "vec", tag = "3")]
        pub copy_id: ::prost::alloc::vec::Vec<u8>,
    }
    #[derive(Clone, PartialEq, Eq, Hash, ::prost::Oneof)]
    pub enum Value {
        #[prost(message, tag = "2")]
        File(File),
        #[prost(bytes, tag = "3")]
        SymlinkId(::prost::alloc::vec::Vec<u8>),
        #[prost(bytes, tag = "4")]
        TreeId(::prost::alloc::vec::Vec<u8>),
    }
}
#[derive(Clone, PartialEq, ::prost::Message)]
pub struct Tree {
    #[prost(message, repeated, tag = "1")]
    pub entries: ::prost::alloc::vec::Vec<tree::Entry>,
}
/// Nested message and enum types in `Tree`.
pub mod tree {
    #[derive(Clone, PartialEq, Eq, Hash, ::prost::Message)]
    pub struct Entry {
        #[prost(string, tag = "1")]
        pub name: ::prost::alloc::string::String,
        #[prost(message, optional, tag = "2")]
        pub value: ::core::option::Option<super::TreeValue>,
    }
}
#[derive(Clone, PartialEq, Eq, Hash, ::prost::Message)]
pub struct Commit {
    #[prost(bytes = "vec", repeated, tag = "1")]
    pub parents: ::prost::alloc::vec::Vec<::prost::alloc::vec::Vec<u8>>,
    #[prost(bytes = "vec", repeated, tag = "2")]
    pub predecessors: ::prost::alloc::vec::Vec<::prost::alloc::vec::Vec<u8>>,
    /// Alternating positive and negative terms
    #[prost(bytes = "vec", repeated, tag = "3")]
    pub root_tree: ::prost::alloc::vec::Vec<::prost::alloc::vec::Vec<u8>>,
    /// Labels for the terms of a conflict.
    #[prost(string, repeated, tag = "10")]
    pub conflict_labels: ::prost::alloc::vec::Vec<::prost::alloc::string::String>,
    #[prost(bytes = "vec", tag = "4")]
    pub change_id: ::prost::alloc::vec::Vec<u8>,
    #[prost(string, tag = "5")]
    pub description: ::prost::alloc::string::String,
    #[prost(message, optional, tag = "6")]
    pub author: ::core::option::Option<commit::Signature>,
    #[prost(message, optional, tag = "7")]
    pub committer: ::core::option::Option<commit::Signature>,
    #[prost(bytes = "vec", optional, tag = "9")]
    pub secure_sig: ::core::option::Option<::prost::alloc::vec::Vec<u8>>,
}
/// Nested message and enum types in `Commit`.
pub mod commit {
    #[derive(Clone, Copy, PartialEq, Eq, Hash, ::prost::Message)]
    pub struct Timestamp {
        #[prost(int64, tag = "1")]
        pub millis_since_epoch: i64,
        #[prost(int32, tag = "2")]
        pub tz_offset: i32,
    }
    #[derive(Clone, PartialEq, Eq, Hash, ::prost::Message)]
    pub struct Signature {
        #[prost(string, tag = "1")]
        pub name: ::prost::alloc::string::String,
        #[prost(string, tag = "2")]
        pub email: ::prost::alloc::string::String,
        #[prost(message, optional, tag = "3")]
        pub timestamp: ::core::option::Option<Timestamp>,
    }
}
